/*******************************************
 *                                         *
 *      Lempel-Ziv 77 Data Compressor      *
 *      Demo Program by Rachad ALAO        *
 *      ENST / Telecom Copyright 1997      *
 *                                         *
 *******************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

struct Match {
        int Length;
        int Pointer;
        };


/* look ahead buffer size */
int N = 512;


/* look ahead buffer size in bits */
int NB = 9;


/* input words buffer size */
int M = 32;


/* Words Buffer size in bits */
int MB = 5;


/* The Words Buffer and Base of the Allocated memory */
unsigned char *Buffer, *Base;
/* Number of words in the Word Buffer, used to handle the eof */
int           NW = 0;

/* The ascii Probabilities Board */
unsigned long Proba[256];


/***********************************************************************
                         My Bit File library
 ***********************************************************************/

#define Cache_Size 20000
#define Error -1
#define Ok 1

struct BFILE {
        FILE *f;                          /* The physical file                   */
        unsigned char Cache[Cache_Size];  /* The Cache used to acces the File    */
        unsigned char CByte;              /* The Current Word being written      */
        int  CBPos;                       /* Current Pos. in the Current Word    */
        unsigned int CPos;                /* Current Pos. in the Cache           */
        unsigned int CEnd;                /* Cache End if not full               */
        char Mode;                        /* R = Read, W = Write                 */
        char eof;                         /* !=0 if end of file                  */
        };


/**** Opens a Bit File ****/
struct BFILE *BF_Open ( const char *Name, const char Mode)
{ struct BFILE *Tmp;
  char FMode[3];

Tmp = (struct BFILE * )malloc(sizeof(struct BFILE));
if (!Tmp)
   return NULL;

FMode[0] = Mode;
FMode[1] = 'b';
FMode[2] = '\0';

Tmp->f = fopen(Name, FMode);

if (Tmp->f) {
   Tmp->Mode = Mode;

   if (Mode == 'r' ) {
      /* Fills the bit file cache if in R mode*/
      Tmp->CEnd  = fread (Tmp->Cache, sizeof(unsigned char), Cache_Size, Tmp->f);
      Tmp->CByte = Tmp->Cache[0];
      Tmp->CBPos = 7;
      Tmp->CPos  = 1;
      Tmp->eof   = 0;
      }

   else {
      Tmp->CEnd  = 0;
      Tmp->CPos  = 0;
      Tmp->CBPos = 0;
      Tmp->CByte = 0;
      Tmp->eof   = 1;
      }
   }
else {
     free (Tmp);
     return NULL;
     }

return Tmp;

}




/**** Writes nbits bits in the BFILE *f ****/
int BF_Write ( unsigned long bits, int nbits, struct BFILE *f)
{ int i = 0;

if (!f || f->Mode != 'w')
   return Error;

if (nbits > 8 * sizeof(unsigned long))
   nbits = 8 * sizeof (unsigned long);

for (i = nbits-1; i>=0; i--) {
    f->CByte = (f->CByte<<1) + ( (bits>>i) & 1 );
    f->CBPos ++;
    if (f->CBPos == 8) {
       f->Cache[f->CPos] = f->CByte;
       f->CPos++;
       if (f->CPos == Cache_Size) {
          fwrite(f->Cache, sizeof(unsigned char), Cache_Size, f->f);
          f->CPos = 0;
          }
       f->CByte = 0;
       f->CBPos = 0;
       }
    }

return Ok;

}



/**** Reads nbits bits in the BFILE *f ****/
unsigned long BF_Read ( int nbits, struct BFILE *f)
{unsigned long bits = 0;
 int   i    = 0;

if (!f || f->Mode != 'r')
   return 0;

if (nbits >8 * sizeof(unsigned long) )
   nbits = 8 * sizeof (unsigned long);

for (i = nbits-1; i>=0; i--) {
    bits = (bits<<1) + ( (f->CByte>>f->CBPos) & 1 );
    f->CBPos --;
    if (f->CBPos == -1 && !f->eof) {
       f->CByte = f->Cache[f->CPos] ;
       f->CPos++;
       if (f->CPos == f->CEnd) {
          f->CEnd = fread(f->Cache, sizeof(unsigned char), Cache_Size, f->f);
          if (!f->CEnd)
             f->eof = 1;
          f->CPos = 0;
          }
 //      f->CByte = f->Cache[0];
       f->CBPos = 7;
       }
    }

return bits;
}



/**** Close a BFILE *f ****/
int BF_Close ( struct BFILE *f)
{

if (!f)
   return Error;

if (f->Mode == 'w') {
   if (f->CBPos != 0)
      BF_Write(0, 8 - f->CBPos, f);
   if (f->CPos != 0)
      fwrite(f->Cache, sizeof(unsigned char), f->CPos, f->f);
   }

fclose (f->f);

f->f = NULL;
free(f);
f = NULL;

return Ok;

}

/***********************************************************************
                  End of my Bit File library
 ***********************************************************************/


/* The Input File */
struct BFILE  *Input;
unsigned long I_Length = 0;


/* The Output File */
struct BFILE  *Output;
unsigned long O_Length = 0;


/** Inits the Buffer **/
void InitB(void)
{ int i = 0;
  /* We allow twice of the necessary mem for speed optimasation purposes */
  Base   = (unsigned char *)malloc( 2*(N+M)*sizeof(char) );
  Buffer = Base;
  if (!Buffer) {
   printf("lz77 : Memory allocation error. Exiting...\n");
   exit(1);
   }
  for (i=0; i<2*(N+M); i++)
      Buffer[i] = 0;
  NW = 0;
  for (i=0; i<256; i++)
      Proba[i] = 0;
}


/** Closes the Buffer **/
void CloseB(void)
{
 free(Base);
}


/** Shifts the Buffer to the left and feeds it by the right **/
void C_Shift(int n)
{int  i     = 0;
 static char           D_eof  = 0;      /* EOF indicator */
 static unsigned int   Offset = 0;      /* Buffer Offset Pos to Base */
 unsigned char         tmp = 0;

if (!n)
   return;

/* shifts the buffer: optimised version */
if ( Offset + n < N+M ) {
   Offset += n;
   Buffer = (unsigned char *)(Base + Offset);
   }
else {
   Offset = 0;
   for (i=0; i<N+M-n; i++)
       Base[i] = Buffer[i+n];
   Buffer = Base;
   }

/* feeds the buffer */
while (NW < M && !D_eof ) {
      D_eof = Input->eof;
      tmp = (unsigned char)BF_Read(8, Input);
      Buffer[N+NW] = tmp;
      Proba[tmp]++;
      I_Length += 8;
      NW++;
      }

}



/** Outputs the lookahead buffer words to output file and shifts the Buffer to the left  **/
void D_Shift(int n)
{int  i     = 0;
 static unsigned int   Offset = 0;      /* Buffer Offset Pos to Base */
 unsigned char         tmp = 0;

if (!n)
   return;

/* outputs n words to the output file */
for (i=0; i<n; i++) {
    tmp = Buffer[N+i];
    BF_Write((unsigned long)tmp, 8, Output);
    Proba[tmp]++;
    O_Length += 8;
    NW--;
    }

/* shifts the buffer: optimised version */
if ( Offset + n < N+M ) {
   Offset += n;
   Buffer = (unsigned char *)(Base + Offset);
   }
else {
   Offset = 0;
   for (i=0; i<N+M-n; i++)
       Base[i] = Buffer[i+n];
   Buffer = Base;
   }

}




/** Searchs for the longest matching string **/
struct Match Search(void)
{ int   i  = 0;
  int   j  = 0;
  int   ml = 0; /* indicates the current max length */
  int   mp = 0; /* indicates the max matching string pointer */
  struct Match tmp;

  while ( i<N && ml<NW) {

        j = 0;

        while (j<NW && Buffer[i+j] == Buffer[N+j])
              j++;

        if (j>ml) {
           ml = j;
           mp = i;
           }

        i++;

        }

  tmp.Length  = ml;
  tmp.Pointer = mp;
  return tmp;

}




/** Compress the Input File to an Output File */
void Compress(void)
{ struct Match  match;

NW = 0;
C_Shift(M);
while (NW > 0 ) {
      match = Search();

      if (match.Length < 2) {  /* length>2 we output the ascii char */
         BF_Write(1, 1, Output);
         BF_Write((unsigned long)Buffer[N], 8, Output);
         O_Length += (8 + 1);
         NW--;
         C_Shift (1);

         }
      else {   /* else we output a coded word (pointer, length - 1) */
           BF_Write(0, 1, Output);
           BF_Write((unsigned long)(match.Length - 1), MB, Output);
           BF_Write((unsigned long)match.Pointer, NB, Output);
           O_Length += (MB + NB + 1);
           NW    -= match.Length;
           C_Shift (match.Length);
           }

      }
I_Length = I_Length/8;
O_Length = O_Length/8;

}



/** Decompress the Input File to an Output File */
void Decompress(void)
{ unsigned char Src     = 0;  /* Source indicator */
  int           Pointer = 0;
  int           Length  = 0;
  int           i       = 0;
/*  int           c       = 0;   decompression cycle */

NW = 0;
while (!Input->eof) {

      Src   = (unsigned char)BF_Read(1, Input);

      if (Src == 1) {  /* we read an immediate ascii code */
         Buffer[N] = (unsigned char)BF_Read(8, Input);
         I_Length += (8 + 1);
         NW++;
         D_Shift (1);
         }
      else {   /* else we read a coded word (pointer, length) */
           Length   = (int)BF_Read(MB, Input) + 1;
           Pointer  = (int)BF_Read(NB, Input);
           for (i=0; i<Length; i++)
               Buffer[N+i] = Buffer[Pointer+i];
           I_Length += (MB + NB + 1);
           NW       += Length;
           D_Shift  (Length);
           }

      }
I_Length = I_Length/8;
O_Length = O_Length/8;

}



/* just returns in Ext the ext filename */
void Ext(char *_ext, const char *_filename)
{ int  i   = 0;
  int  pos = strlen(_filename) -1;

while (pos>=0 && _filename[pos]!='.')
      pos--;
if (pos>=0) {
   for (i=pos+1; i<strlen(_filename); i++)
       _ext[i-pos-1] = _filename[i];
   _ext[strlen(_filename)-pos-1] = '\0';
   }
else
    _ext[0] = '\0';
}



/* just returns in Name the name filename */
void Name(char *_name, const char *_filename)
{ int  i   = 0;
  int  pos = strlen(_filename) -1;

while (pos>=0 && _filename[pos]!='.')
      pos--;
if (pos < 0) pos = strlen(_filename);
if (pos) {
   for (i=0; i<pos; i++)
       _name[i] = _filename[i];
   _name[pos] = '\0';
   }
else
    _name[0] = '\0';

}





/**** Build a report file ****/
void c_report(char *_name)
{ FILE *f;
  double Entropy = 0.0;
  double P[256];
  int i = 0;
  char tmp[256];

Name(tmp, _name);
strcat(tmp, ".rpt");
f = fopen(tmp,"r");
if (f) {
   Name(tmp, _name);
   strcat(tmp, ".rp1");
   }
f = fopen(tmp,"wt");
if (!f) {
   printf("lz77 : I/O error. Exiting...\n");
   exit(1);
   }

fprintf(f,"\t\t**** LZ77 v.1.0 report ****\n\n");

fprintf (f,"-> compressing %s\n", _name);

fprintf (f,"Buffer Size : %d, Look ahead Buffer Size : %d\n"
         , N, M);

fprintf (f,"Out Size : %ld, In Size : %ld, Ratio : %d %% \n"
         , O_Length, I_Length, (100*O_Length)/I_Length);

fprintf (f,"-> file compressed successfuly!\n\n");

fprintf (f,"\n** Source Statistics :\n\n");
/**** Calculation of Input Entropy as iid source ****/
for (i=0; i<256; i++) {
    P[i]     =( (double)(Proba[i]) / (double)I_Length);
    if (P[i] > 1.0e-9)
       Entropy += -P[i] * log (P[i]) / log(2);
    else P[i] = 0.0;
    }
fprintf(f,"  Theoritical Entropy ( if Input = iid ) = %f, Reached Entropy = %f\n\n"
        ,(float)Entropy, (8.0*O_Length)/I_Length );
/**** Printing the ascii Distribution ****/
fprintf(f,"  Ascii Distribution :\n");
for (i=0; i<256; i++) {
    if ( i<31 )
       fprintf(f,"  Ascii(%d) = . Occured %ld => P[%d] = %f\n"
               , i, Proba[i], i, (float)P[i] );
    else
       fprintf(f,"  Ascii(%d) = %c Occured %ld => P[%d] = %f\n"
               , i, i, Proba[i], i, (float)P[i] );
    }

fprintf(f,"End of LZ77 Report.\n");
fclose(f);

}



/**** Build a report file ****/
void d_report(char *_name)
{ FILE *f;
  double Entropy = 0.0;
  double P[256];
  int i = 0;
  char tmp[256];

Name(tmp, _name);
strcat(tmp, ".rpt");
f = fopen(tmp,"r");
if (f) {
   Name(tmp, _name);
   strcat(tmp, ".rp1");
   }
f = fopen(tmp,"wt");
if (!f) {
   printf("lz77 : I/O error. Exiting...\n");
   exit(1);
   }

fprintf(f,"\t\t**** LZ77 v.1.0 report ****\n\n");

fprintf (f,"-> decompressing %s\n", _name);

fprintf (f,"Buffer Size : %d, Look ahead Buffer Size : %d\n"
         , N, M);

fprintf (f,"Out Size : %ld, In Size : %ld, Ratio : %d %% \n"
         , O_Length, I_Length, (100*O_Length)/I_Length);

fprintf (f,"-> file decompressed successfuly!\n\n");

fprintf (f,"\n** Source Statistics :\n\n");
/**** Calculation of Input Entropy as iid source ****/
for (i=0; i<256; i++) {
    P[i]     =( (double)(Proba[i]) / (double)O_Length);
    if (P[i] > 1.0e-9)
       Entropy += -P[i] * log (P[i]) / log(2);
    else P[i] = 0.0;
    }
fprintf(f,"  Theoritical Entropy ( if Output = iid ) = %f, Reached Entropy = %f\n\n"
        ,(float)Entropy, (8.0*I_Length)/O_Length );
/**** Printing the ascii Distribution ****/
fprintf(f,"  Ascii Distribution :\n");
for (i=0; i<256; i++) {
    if ( i<31 )
       fprintf(f,"  Ascii(%d) = . Occured %ld => P[%d] = %f\n"
               , i, Proba[i], i, (float)P[i] );
    else
       fprintf(f,"  Ascii(%d) = %c Occured %ld => P[%d] = %f\n"
               , i, i, Proba[i], i, (float)P[i] );
    }

fprintf(f,"End of LZ77 Report.\n");
fclose(f);

}



/* just returns 2^n */
int _2p(int n)
{ int i;
  int tmp = 1;
for  (i=0; i<n; i++)
     tmp *= 2;
return tmp;
}



/* prints a help message */
void Help(void)
{
   printf("\n LZ77 v.1.0 by Rachad ALAO ( ENST / Telecom Paris). usage : \n\n");
   printf("    lz77 -? -c -help -d -r -b size   <Input> <optionnal Output>\n\n");
   printf("        <Input>  : Input file name\n");
   printf("        <Output> : Optionnal Output file name\n");
   printf("        -?       : displays this help message\n");
   printf("        -c       : indicates lz77 to perform compress command (default)\n");
   printf("        -help    : displays this help message\n");
   printf("        -d       : indicates lz77 to perform decompress command\n");
   printf("        -r       : indicates lz77 to generate a report file\n");
   printf("        -b size  : indicates lz77 the size of the buffer to use in bits\n");
   printf("                   default size is 9 => 512 bytes buffer size\n");
   exit(1);
}




void main (int argc, char **argv)
{int   i = 0;
 int   pos = 0;           /* parsing Position */
 int   In_Ok = 0;         /* indicates if input file name is already read */
 int   Out_Ok = 0;        /* indicates if output file name is already read */
 int   Report = 0;        /* Report flag */
 char  In_Name[256];      /* input Name   */
 char  In_Ext[40];        /* input Extension */
 char  In_Filename[256];  /* The entier input file name */
 char  Out_Ext[40];       /* output Extension */
 char  Out_Filename[256]; /* The entier output file name */
 char  Cmd = 'c';         /* Command to perform c=compress, d=decompress, etc. */
 char  Header[5];         /* Header of the Input File */
 int   tmp = 0;

if ( argc < 2 ) Help();

/* parses the command line */
for (pos=1; pos<argc; pos++) {
    if (!strcmp(argv[pos], "-c"))
       Cmd = 'c';
    else if (!strcmp(argv[pos], "-d") || !strcmp(argv[pos], "-D"))
            Cmd = 'd';
    else if (!strcmp(argv[pos], "-r") || !strcmp(argv[pos], "-R"))
            Report = 1;
    else if (!strcmp(argv[pos], "-b") || !strcmp(argv[pos], "-B")) {
            NB = atoi(argv[pos+1]);
            if (NB<6)  NB = 6;
            if (NB>13) NB = 13;
            N  = _2p(NB);
            pos++;
            }
    else if (!strcmp(argv[pos], "-help") || !strcmp(argv[pos], "-?"))
            Help();
    else if (!In_Ok) {
            Name(In_Name, argv[pos]);
            Ext(In_Ext, argv[pos]);
            strcpy(In_Filename, argv[pos]);
            In_Ok   = 1;
            }
    else if (In_Ok && !Out_Ok) {
            strcpy(Out_Filename, argv[pos]);
            Out_Ok   = 1;
            }
    else Help();
    }

printf("\n LZ77 v.1.0 by Rachad ALAO ( ENST / Telecom Paris). \n\n");

/* opens the input file  */
Input  = BF_Open(In_Filename,'r');
if (!Input && Cmd == 'd') {
      strcat(In_Filename, ".lz");
      Input  = BF_Open(In_Filename,'r');
      }
if (!Input) {
   printf("lz77 : I/O error. Exiting...\n");
   exit(1);
   }


if (Cmd == 'c') {
   /* Opens the output filename  */
   if (!Out_Ok) {
      strcpy(Out_Filename, In_Name);
      strcat(Out_Filename,".lz");
      }
   Output = BF_Open(Out_Filename, 'w');
   if (!Output) {
      printf("lz77 : I/O error. Exiting...\n");
      exit(1);
      }

   /* Adds the Lz77 Header to the output file */
   BF_Write('L', 8, Output);
   BF_Write('Z', 8, Output);
   BF_Write('7', 8, Output);
   BF_Write('7', 8, Output);
   BF_Write((unsigned long)NB, 8, Output);
   BF_Write((unsigned long)strlen(In_Ext), 8, Output);
   for (i=0; i<strlen(In_Ext); i++)
       BF_Write((unsigned long)In_Ext[i], 8, Output);

   printf  ("-> compressing %s, please wait...\n", In_Filename);
   InitB();
   Compress ();
   CloseB();
   printf  ("-> file compressed successfuly!\n");
   printf  ("-> Compression Ratio : %d / 100\n"
        ,(int)((100*O_Length)/I_Length));
   }
else {
   /* Checks the Header of the input file */
   for (i=0; i<4; i++)
       Header[i] = (char)BF_Read(8, Input);
   Header[4] = '\0';
   if (strcmp(Header,"LZ77")) {
      printf("lz77 : Bad Input file format. Exiting...\n");
      exit(1);
      }
   NB = (int)BF_Read(8, Input);
   N  = _2p(NB);

   /* opens the output file name */
   tmp = (int)BF_Read(8, Input);
   for (i=0; i<tmp; i++)
       Out_Ext[i] = (char)BF_Read(8, Input);
   Out_Ext[tmp] = '\0';
   if (!Out_Ok) {
      strcpy(Out_Filename, In_Name);
      strcat(Out_Filename, ".");
      strcat(Out_Filename, Out_Ext);
      }
   Output = BF_Open(Out_Filename, 'w');
   if (!Output) {
      printf("lz77 : I/O error. Exiting...\n");
      exit(1);
      }

   printf  ("-> decompressing %s, please wait...\n", In_Filename);
   InitB();
   Decompress();
   CloseB();
   printf  ("-> file decompressed successfuly!\n");
   printf  ("-> Decompression Ratio : %d / 100\n"
        ,(int)((100*O_Length)/I_Length));
   }

if (Report)
   if (Cmd=='c') c_report(In_Filename);
   else d_report(In_Filename);

BF_Close (Input);
BF_Close (Output);

}




