/*******************************************
 *                                         *
 *      Lempel-Ziv 77 Data Compressor      *
 *      Demo Program by Rachad ALAO        *
 *      ENST / Telecom Copyright 1997      *
 *                                         *
 *******************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* look ahead buffer size */
#define N 512
/* look ahead buffer size in bits */
#define NB 9

/* input words buffer size */
#define M 32
/* Words Buffer size in bits */
#define MB 5


struct Match {
        int Length;
        int Pointer;
        };

/* The Words Buffer */
unsigned char Buffer[N+M];
/* Number of words in the Word Buffer, used to handle the eof */
int           NW = 0;


/* The Input File */
FILE          *Input;
unsigned long I_Length = 0;

/* The Output File */
FILE          *Output;
unsigned long O_Length = 0;




/** Shifts the Buffer to the left and feeds it by the right **/
void Shift(int n)
{int i   = 0;
 int tmp = 0;

if (!n)
   return;

/* shifts the buffer */
for (i=0; i<N+M-n; i++)
    Buffer[i] = Buffer[i+n];

/* feeds the buffer */
while (NW < M && fscanf(Input,"%d",&tmp)!=EOF ) {
      Buffer[N+NW] = (unsigned char)tmp;
      I_Length += 8;
      NW++;
      }

}


/** Searchs for the longest matching string **/
struct Match Search(void)
{ int   i  = 0;
  int   j  = 0;
  int   ml = 0; /* indicates the current max length */
  int   mp = 0; /* indicates the maw matching string pointer */
  struct Match tmp;

  while ( i<N && ml<NW) {

        j = 0;

        while (j<NW && Buffer[i+j] == Buffer[N+j])
              j++;
        if (j>ml) {
           ml = j;
           mp = i;
           }

        i++;

        }

  tmp.Length  = ml;
  tmp.Pointer = mp;
  return tmp;

}




/** Writes the result of cycle C in the output file **/
void Write_Cycle(int C, struct Match match)
{int  i   = 0;
 int  Bit = 0;
 char tmp[256];

if (!Output)
   return;

strcpy(tmp,"");

if (match.Length < 2) {  /* length>2 we output the ascii char */
   for (i=7; i>=0; i--) {
       Bit = (Buffer[N]>>i) & 1;
       if (Bit) strcat(tmp,"1");
       else     strcat(tmp,"0");
       }
   fprintf(Output,"Cycle %d : 1 %s\n", C, tmp);
   O_Length += (8 + 1);
   }
else {   /* else we output a coded word (pointer, length-1) */
   for (i=MB-1; i>=0; i--) {
       match.Length --;
       Bit = (match.Length>>i) & 1;
       if (Bit) strcat(tmp,"1");
       else     strcat(tmp,"0");
       }
   strcat(tmp," ");
   for (i=NB-1; i>=0; i--) {
       Bit = (match.Pointer>>i) & 1;
       if (Bit) strcat(tmp,"1");
       else     strcat(tmp,"0");
       }
   fprintf(Output,"Cycle %d : 0 %s\n", C, tmp);
   O_Length += (MB + NB + 1);
   }

}




/** Simulates the compression of the Input */
void Compress(void)
{ int    c = 0; /* the cycle number */
  struct Match  match;

NW = 0;
Shift(M);
while (NW > 0 ) { 
      match = Search();
      Write_Cycle(c++, match);
      if (match.Length >=2 ) {
         NW    -= match.Length;
         Shift (match.Length);
         }
      else {
           NW--;
           Shift (1);
           }

      }

}




void main (int argc, char **argv)
{int i=0;

if ( argc < 3 ) {
   printf("\n LZ77 Demo by Rachad ALAO ( ENST / Telecom Paris ). usage : \n");
   printf("\t lz77 <Input file name> <Output file Name>\n");
   exit(1);
   }

printf("\n LZ77 Demo by Rachad ALAO ( ENST / Telecom Paris ). \n");

Input  = fopen(argv[1],"rt");
Output = fopen(argv[2],"wt");

if (!Input || !Output) {
   printf("lz77 : I/O error. Exiting...\n");
   exit(1);
   }

for (i=0; i<N+M; i++)
    Buffer[i] = 0;

printf  ("-> compressing %s, please wait...\n", argv[1]);
fprintf (Output,"-> compressing %s, please wait...\n\n", argv[1]);

Compress ();

fprintf (Output,"\n-> file compressed successfuly!\n");
fprintf (Output,"\n-> Compression Ratio : %d / 100\n"
        ,(int)((100*O_Length)/I_Length));

printf  ("-> file compressed successfuly!\n");
printf ("-> Compression Ratio : %d / 100\n"
        ,(int)((100*O_Length)/I_Length));

fclose (Input);
fclose (Output);

}




