/*******************************************
 *                                         *
 *      Lempel-Ziv Data Compressor         *
 *      Demo Program by Rachad ALAO        *
 *      ENST / Telecom Copyright 1997      *
 *                                         *
 *******************************************/

# include "lza.h"


/**** Init the Words Dico and its H_Table ****/
void Dico_Init(void)
{ uint i = 0;

  Dico = (Word *) malloc(Dico_Size * sizeof (Word));
  if (!Dico) {
     printf("lza : Memory allocation error. exiting...\n");
     exit(1);
     }

  for (i=0; i<Dico_Size; i++) {
      Dico[i].String = NULL;
      Dico[i].Length = 0;
      Dico[i].N      = 0;
      }

  for (i=0; i<HC; i++)
      H_Table[i] = NULL;

  Dico_Pos = 0;

}



/**** Frees the Words Dico and its H_Table ****/
void Dico_Close(void)
{ int    i     = 0;
  H_Node *Next = NULL;
  H_Node *Tmp  = NULL;

  for (i=0; i<Dico_Pos; i++)
      if (Dico[i].String)
         free (Dico[i].String);

  for (i=0; i<HC; i++)
      if (H_Table[i]) {
         Next = H_Table[i]->Next;
         while (Next) {
               Tmp = Next->Next;
               free(Next);
               Next = Tmp;
               }
         free (H_Table[i]);
         }

  free (Dico);

}



/***** Lookup for String in the Dico ****/
int Dico_Lookup (byte *String, uint Length)
{ uint   i = 0;
  uint   j = 0;
  int    Match = 0;
  H_Node *Tmp  = NULL;

if ( String == NULL ) return Error;

Tmp = H_Table[H(String, Length)];

while ( Tmp && !Match ) {
    Match = 0;
    i     = Tmp->Pos;
    if (Dico[i].Length == Length ) {
       Match = 1;      j = 0;
       while  ( j<Dico[i].Length && Match ) {
              if (Dico[i].String[j] != String[j] )
                 Match = 0;
              j++;
              }
       }
    Tmp = Tmp->Next;
    }

if ( !Match )
   return NotFound;

return i;

}


/**** Reads the Word Located at Pos Pos in the Dico ****/
Word Dico_Read (int Pos)
{uint i;
 Word Tmp;
 H_Node *tmp = NULL;

if (Pos < Dico_Pos && Pos >= 0)  {
   Dico[Pos].N  ++;

   i = Pos;
   /* Updates the Pos of the Word in the Dico (sorts it) */
   while (i>0 && Dico[i].N>Dico[i-1].N) {
         Tmp       = Dico[i-1];
         Dico[i-1] = Dico[i];
         Dico[i]   = Tmp;
         /* updates their H_Table entries */
         tmp = H_Table[H(Dico[i].String, Dico[i].Length)];
         while (tmp && tmp->Pos != i-1) {
             tmp = tmp->Next;
             }
         if (tmp)
            tmp->Pos = i;
         tmp = H_Table[H(Dico[i-1].String, Dico[i-1].Length)];
         while (tmp && tmp->Pos != i) {
             tmp = tmp->Next;
             }
         if (tmp)
            tmp->Pos = i-1;

         i --;
         }

   return Dico[Pos];
   }

Tmp.String = NULL;
Tmp.Length = 0;
Tmp.N      = 0;

return Tmp;
}



/**** Writes the String String in the Dico ****/
int Dico_Write (byte *String, uint Length)
{int i     = 0;
 int H_Pos = 0;
 H_Node *Tmp = NULL;
 H_Node *Old = NULL;

if (Dico_Pos >= Dico_Size) {
   /* frees the FBLength less used words */
   for (i=1; i<= FBLength; i++) {
       /* Frees its H_Table entry */
       H_Pos = H(Dico[Dico_Size-i].String, Dico[Dico_Size-i].Length);
       Tmp = H_Table[H_Pos];
       Old = Tmp;
       while (Tmp && Tmp->Pos != Dico_Size - i) {
             Old = Tmp;
             Tmp = Tmp->Next;
             }
       if (Tmp && Tmp->Pos == Dico_Size - i && Tmp != Old)
          Old->Next = Tmp->Next;
       else if (Tmp && Tmp->Pos == Dico_Size - i && Tmp == Old)
          H_Table[H_Pos] = Tmp->Next;
       if (Tmp) {
          free(Tmp);
          Tmp = NULL;
          }
       /* frees its Dico Entry */
       free (Dico[Dico_Size-i].String);
       Dico[Dico_Size-i].String = NULL;
       Dico[Dico_Size-i].Length = 0;
       Dico[Dico_Size-i].N      = 0;

       }
   Dico_Pos -= FBLength;
   }

Dico[Dico_Pos].String = (byte *)malloc(Length*sizeof(byte));
if (!Dico[Dico_Pos].String) {
   printf("lza : Memory Allocation Error. Exiting...\n");
   exit(1);
   }
memcpy(Dico[Dico_Pos].String, String, Length);
Dico[Dico_Pos].N      = 1;
Dico[Dico_Pos].Length = Length;
/* Adds its H_Table entry */
H_Pos = H(String, Length);
Tmp = H_Table[H_Pos];
Old = Tmp;
while (Tmp) {
      Old = Tmp;
      Tmp = Tmp->Next;
      }
if ( Tmp != Old) {
   Old->Next = (H_Node *)malloc(sizeof(H_Node));
   Tmp       = Old->Next;
   }
else {
   H_Table[H_Pos] = (H_Node *)malloc(sizeof(H_Node));
   Tmp            = H_Table[H_Pos];
   }
if (!Tmp) {
   printf("lza : Memory Allocation Error. Exiting...\n");
   exit(1);
   }
Tmp->Pos  = Dico_Pos;
Tmp->Next = NULL;


Dico_Pos ++;

return (int)(Dico_Pos - 1);
}



/**** Calculate the H function for String ****/
int H(byte *String, uint Length)
{ int i;
  long tmp = 0;

for (i=0; i<Length; i++)
    tmp += String[i];

return ( (int) (tmp - HC*tmp/HC ) );
}





/**** Opens a Bit File ****/
BFILE *BF_Open ( const char *Name, const char Mode)
{ BFILE *Tmp;
  char FMode[3];

Tmp = (BFILE * )malloc(sizeof(BFILE));
if (!Tmp)
   return NULL;

FMode[0] = Mode;
FMode[1] = 'b';
FMode[2] = '\0';

Tmp->f = fopen(Name, FMode);

if (Tmp->f) {
   Tmp->Mode = Mode;

   if (Mode == 'r' ) {
      /* Fills the bit file cache if in R mode*/
      Tmp->CEnd  = fread (Tmp->Cache, sizeof(byte), Cache_Size, Tmp->f);
      Tmp->CByte = Tmp->Cache[0];
      Tmp->CBPos = 7;
      Tmp->CPos  = 1;
      Tmp->eof   = 0;
      }

   else {
      Tmp->CEnd  = 0;
      Tmp->CPos  = 0;
      Tmp->CBPos = 0;
      Tmp->CByte = 0;
      Tmp->eof   = 1;
      }
   }
else
    return NULL;

return Tmp;

}




/**** Writes nbits bits in the BFILE *f ****/
int BF_Write ( ulong bits, int nbits, BFILE *f)
{ int i = 0;

if (!f || f->Mode != 'w')
   return Error;

if (nbits > 8 * sizeof(ulong))
   nbits = 8 * sizeof (ulong);

for (i = nbits-1; i>=0; i--) {
    f->CByte = (f->CByte<<1) + ( (bits>>i) & 1 );
    f->CBPos ++;
    if (f->CBPos == 8) {
       f->Cache[f->CPos] = f->CByte;
       f->CPos++;
       if (f->CPos == Cache_Size) {
          fwrite(f->Cache, sizeof(byte), Cache_Size, f->f);
          f->CPos = 0;
          }
       f->CByte = 0;
       f->CBPos = 0;
       }
    }

return Ok;

}



/**** Reads nbits bits in the BFILE *f ****/
ulong BF_Read ( int nbits, BFILE *f)
{ulong bits = 0;
 int   i    = 0;

if (!f || f->Mode != 'r'|| f->eof)
   return 0;

if (nbits >8 * sizeof(ulong) )
   nbits = 8 * sizeof (ulong);

for (i = nbits-1; i>=0; i--) {
    bits = (bits<<1) + ( (f->CByte>>f->CBPos) & 1 );
    f->CBPos --;
    if (f->CBPos == -1) {
       f->CByte = f->Cache[f->CPos] ;
       f->CPos++;
       if (f->CPos == f->CEnd) {
          f->CEnd = fread(f->Cache, sizeof(byte), Cache_Size, f->f);
          if (!f->CEnd) {
             f->eof = 1;
             return bits;
             }
          f->CPos = 0;
          }
 //      f->CByte = f->Cache[0];
       f->CBPos = 7;
       }
    }

return bits;
}



/**** Close a BFILE *f ****/
int BF_Close ( BFILE *f)
{

if (!f)
   return Error;

if (f->Mode == 'w') {
   if (f->CPos != 0)
      fwrite(f->Cache, sizeof(byte), f->CPos, f->f);
   }

fclose (f->f);

f->f = NULL;
free(f);
f = NULL;

return Ok;

}



/**** Compress Input to Output ****/
void Compress (BFILE *Input, BFILE *Output)
{ byte Buffer[1*K];       /* Buffer of strings     */
  uint Pos   = 0;         /* Pos in the Buffer     */
  uint BDPos = 0;         /* Pos of Buffer in the Dico */
  byte CByte;             /* Current Byte       */
  int  Lookup = NotFound;
  int  Delayed_EOF = 0;

/* Output the first char */
CByte = BF_Read(8, Input);
In_Size++;
Proba[CByte] ++;

Buffer[Pos] = CByte;
Pos++;

while ( !Delayed_EOF ) {
      Delayed_EOF = Input->eof; /* Receives f->eof with one cycle of delay */
      CByte = BF_Read(8, Input);
      In_Size++;
      Proba[CByte] ++;

      Buffer[Pos] = CByte;
      Pos++;

      Lookup = Dico_Lookup(Buffer, Pos);

      if (Lookup == Error)
         return;

      if (Lookup == NotFound) {
         Pos--;
         if (Pos > 1) { /* Pos > 1 => we output a pointer to Buffer */
            BDPos = Dico_Lookup(Buffer, Pos);
            Dico_Read(BDPos);
            BF_Write(1, 1, Output);      /* 1 Indicates a Dico Pointer */
            BF_Write(BDPos, DICO_BL, Output);
            Out_Size += (DICO_BL + 1);
            }
         else if (Pos == 1) { /* Pos == 1 => we output the ascii char */
            BF_Write(0, 1, Output);      /* 0 Indicates a direct ascii output */
            BF_Write(Buffer[0], 8, Output);
            Out_Size += 9;
            }

         Pos++;
         Dico_Write(Buffer, Pos);
         Buffer[0] = CByte;
         Pos = 1;
         }

      }
Out_Size = Out_Size / 8 + 1;
printf ("Out Size : %ld, In Size : %ld, Ratio : %d %% \n"
        , Out_Size, In_Size, (100*Out_Size)/In_Size);
}




/**** Decompress Input to Output ****/
void Decompress (BFILE *Input, BFILE *Output)
{ uint BDPos = 0;         /* Pos of Buffer in the Dico */
  byte CByte;             /* Current Byte       */
  int  Delayed_EOF = 0;
  int  Org = 0;
  int  i = 0;
  Word Tmp;
  byte Buffer[1*K];       /* Buffer of strings     */
  uint Pos   = 0;         /* Pos in the Buffer     */
  int  Lookup = NotFound;

while ( !Delayed_EOF ) {
      Delayed_EOF = Input->eof; /* Receives f->eof with one cycle of delay */
      Org = (int)BF_Read(1, Input);
      In_Size++;
      if (!Org) { /* means that we read a direct ascii code */
         CByte = BF_Read(8, Input);
         In_Size += 8;
         Proba[CByte] ++;
         BF_Write(CByte, 8, Output);
         Output++;
         /* updates the dico */
         Buffer[Pos] = CByte;
         Pos ++;
         if (Pos > 1) {
            Lookup = Dico_Lookup(Buffer, Pos);
            if (Lookup == Error)
               return;
            if (Lookup == NotFound) {
               Dico_Write(Buffer, Pos);
               Buffer[0] = CByte;
               Pos = 1;
               }
            }
         }
      else { /* else we read a dico entry pos */
         BDPos = (int) BF_Read(DICO_BL, Input);
         In_Size += DICO_BL;
         Tmp = Dico_Read(BDPos);
         for (i=0; i<Tmp.Length; i++) {
             Proba[Tmp.String[i]] ++;
             BF_Write(Tmp.String[i], 8, Output);
             Output++;
             }
         /* updates the dico */
         Buffer[Pos] = Tmp.String[0];
         Pos ++;
         if (Pos > 1) {
            Lookup = Dico_Lookup(Buffer, Pos);
            if (Lookup == Error)
               return;
            if (Lookup == NotFound) {
               Dico_Write(Buffer, Pos);
               for (i=0; i<Tmp.Length; i++)
                   Buffer[i] = Tmp.String[i];
               Pos = Tmp.Length;
               }
            }
         }

      }

In_Size = In_Size / 8 + 1;
printf ("In Size : %ld, Out Size : %ld, Ratio : %d %% \n"
        , Out_Size, In_Size, (100*In_Size)/Out_Size);
}





/**** Build a report file ****/
void report(int Cmd, char *Name)
{ FILE *f;
  double Entropy = 0.0;
  double P[256];
  int i = 0;
  int j = 0;
  char tmp[1*K];
  ulong Size = 0;
/*  H_Node *Tmp = NULL; */

f = fopen("lza.rpt","wt");
if (!f) {
   printf("lza : I/O error. Exiting...\n");
   exit(1);
   }

fprintf(f,"\t\t**** LZA v.1.0 report ****\n\n");

if (Cmd == 'd') {
   fprintf (f,"-> decompressing %s\n", Name);
   Size     = In_Size;
   In_Size  = Out_Size;
   Out_Size = Size;
   }
else
   fprintf (f,"-> compressing %s\n", Name);

fprintf (f,"Out Size : %ld, In Size : %ld, Ratio : %d %% \n"
         , Out_Size, In_Size, (100*Out_Size)/In_Size);

if (Cmd == 'd')
   fprintf (f,"-> file decompressed successfuly!\n\n");
else
   fprintf (f,"-> file compressed successfuly!\n\n");

fprintf (f,"\n** Source Statistics :\n\n");
/**** Calculation of Input Entropy as iid source ****/
for (i=0; i<256; i++) {
    P[i]     =( (double)(Proba[i]) / (double)In_Size);
    if (P[i] > 1.0e-9)
       Entropy += -P[i] * log (P[i]) / log(2);
    else P[i] = 0.0;
    }
fprintf(f,"  Theoritical Entropy ( if Input = iid ) = %f, Reached Entropy = %f\n\n"
        ,(float)Entropy, (8.0*Out_Size)/In_Size );
/**** Printing the ascii Distribution ****/
fprintf(f,"  Ascii Distribution :\n");
for (i=0; i<256; i++) {
    if ( i<31 )
       fprintf(f,"  Ascii(%d) = . Occured %ld => P[%d] = %f\n"
               , i, Proba[i], i, (float)P[i] );
    else
       fprintf(f,"  Ascii(%d) = %c Occured %ld => P[%d] = %f\n"
               , i, i, Proba[i], i, (float)P[i] );
    }
/**** Printing the word Distribution ****/
fprintf(f,"\n  Word Distribution : \n");
for (i=0; i<Dico_Pos; i++) {
    for (j=0; j<Dico[i].Length; j++) {
        if ( Dico[i].String[j] < 31)
           tmp[j] = '.';
        else
           tmp[j] = Dico[i].String[j];
        }

    tmp[Dico[i].Length]='\0';
    fprintf(f,"Pos %d :  Word %s Occured %ld\n", i, tmp, Dico[i].N);
    /** Checks for H_Table Integrity
    Tmp = H_Table[H(Dico[i].String, Dico[i].Length)];
    while (Tmp) {
          fprintf(f, "%d ", Tmp->Pos);
          Tmp = Tmp->Next;
          }
    fprintf(f, "\n");                **/
    }

fprintf(f,"End of LZA Report.\n");
fclose(f);
}



/* just returns 2^n */
int _2p(int n)
{ int i;
  int tmp = 1;
for  (i=0; i<n; i++)
     tmp *= 2;
return tmp;
}



void main (int argc, char **argv)
{int   i = 0;             /* loop counter */
 int   pos = 0;           /* parsing Position */
 int   In_Ok = 0;         /* indicates if input file name is already read */
 int   Report = 0;         /* Report flag */
 char  In_Name[256];      /* input Name   */
 char  Out_Name[256];     /* output name  */
 char  Cmd;               /* Command to perform c=compress, d=decompress, etc. */
 char  Header[5];         /* Header string */
 BFILE *In;               /* Input File descriptor  */
 BFILE *Out;              /* Output File descriptor */


if ( argc < 3 ) {
   printf("\n LZA v.1.0 usage : \n\n");
   printf("    lza -c -d -r -s size   <Input file name> <Output file Name>\n\n");
   printf("        -c       : indicates lza to perform compress command (default)\n");
   printf("        -d       : indicates lza to perform decompress command\n");
   printf("        -r       : indicates lza to generate a report file\n");
   printf("        -s size  : indicates lza the size of the buffer to use in bits\n");
   printf("                   default size is 11 => 2K buffer size\n");
   exit(1);
   }

/* Sets up the Global Default Constants */
DICO_BL   = 11;
Dico_Size = 2*K;
Cmd       = 'c';

/* parses the command line */
for (pos=1; pos<argc; pos++) {
    if (!strcmp(argv[pos], "-c"))
       Cmd = 'c';
    else if (!strcmp(argv[pos], "-d"))
            Cmd = 'd';
    else if (!strcmp(argv[pos], "-r"))
            Report = 1;
    else if (!strcmp(argv[pos], "-s")) {
         DICO_BL = atoi(argv[pos+1]);
         pos++;
         }
    else if (!In_Ok) {
          strcpy (In_Name, argv[pos]);
          In_Ok   = 1;
          }
    else strcpy(Out_Name, argv[pos]);
    }


Dico_Size = _2p(DICO_BL);

In  = BF_Open(In_Name,'r');
Out = BF_Open(Out_Name,'w');

if (!In || !Out) {
   printf("lza : I/O error. Exiting...\n");
   exit(1);
   }


/* Initializes  the Global Tools */
for (i=0; i<256; i++)
    Proba[i] = 0;

In_Size   = 0;
Out_Size  = 0;

/* Compress or Decompress the Input File to the Output File */
if (Cmd == 'd') {
   /* Reads Input File  Buffer */
   Header[0] = (char)BF_Read(8, In);
   Header[1] = (char)BF_Read(8, In);
   Header[2] = (char)BF_Read(8, In);
   Header[3] =  '\0';
   if ( strcmp(Header,"LZA")) {
      printf("lza : Bad Input   file format. Exiting...\n");
      exit(1);
      }
   DICO_BL   = BF_Read(8, In);
   Dico_Size = _2p(DICO_BL);

   printf("-> decompressing %s, please wait...\n", In_Name);
   Dico_Init();
   Decompress(In, Out);
   printf("-> file decompressed successfuly!\n");
   }
else {
   /* Writes Output file header */
   BF_Write('L', 8, Out);
   BF_Write('Z', 8, Out);
   BF_Write('A', 8, Out);
   BF_Write(DICO_BL, 8, Out);

   printf("-> compressing %s, please wait...\n", In_Name);
   Dico_Init();
   Compress (In, Out);
   printf("-> file compressed successfuly!\n");
   }

/* Generates a report file */
if (Report)
   report(Cmd, In_Name);

Dico_Close();

BF_Close(In);
BF_Close(Out);

}







