/*******************************************
 *                                         *
 *      Lempel-Ziv Data Compressor         *
 *      Demo Program by Rachad ALAO        *
 *      ENST / Telecom Copyright 1997      *
 *                                         *
 *******************************************/

# ifndef LZA_H

# define LZA_H
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <ctype.h>
# include <math.h>

# define Ok 1
# define Error -1
# define NotFound -2
# define K 1024
# define Cache_Size 10*K
# define FBLength 2*K
# define HC 547

typedef  unsigned long ulong;
typedef  unsigned int  uint;
typedef  unsigned char byte;



/************************************
 **** Dico Management primitives ****
 ************************************/

typedef struct Word {
       byte  *String;    /* String of bytes contained in the word */
       uint  Length;     /* Size of the String                    */
       ulong N;          /* Number of Occurence                   */
       };

Word *Dico;             /* Dictionnary of the already encoutered words */

uint Dico_Size;         /* Size of the Dictionnary    */

uint Dico_Pos;           /* Position in the Dictionnary */


/** Dico_Init : Inits the strings of the Dico
 **             Input  : void
 **             Output : void.
 **/
void Dico_Init (void);


/** Dico_Lookup : Looks if String is in the Dico
 **               Input  : byte *String, uint Length (the Length of the String)
 **               Output : int Pos of the String if found, NotFound otherwise.
 **/
int Dico_Lookup (byte *String, uint Length);

/** Dico_Read : Reads The String at Pos Pos in the Dico
 **             Input  : The Pos to read
 **             Output : The Word Read or a zero length word if error.
 **/
Word Dico_Read (int Pos);


/** Dico_Write : Writes The String String in the Dico
 **              Input  : The String to write and its length
 **              Output : Error if an error occured.
 **/
int Dico_Write (byte *String, uint Length);


/** Dico_Close : frees the strings of the Dico
 **             Input  : void
 **             Output : void.
 **/
void Dico_Close (void);



/*********************************
 **** Dico Speedup primitives ****
 *********************************/
 typedef struct H_Node {
         int    Pos;      /* Pos in Dico */
         H_Node *Next;    /* Pointer to Next H_Node */
         };

 H_Node *H_Table[HC];     /* The translation H_Table */

 int H(byte *String, uint Length);



/****************************************
 **** Bit File Management primitives ****
 ****************************************/

typedef struct BFILE {
        FILE *f;                 /* The physical file                   */
        byte Cache[Cache_Size];  /* The Cache used to acces the File    */
        byte CByte;              /* The Current Word being written      */
        int  CBPos;              /* Current Pos. in the Current Word    */
        uint CPos;               /* Current Pos. in the Cache           */
        uint CEnd;               /* Cache End if not full               */
        char Mode;               /* R = Read, W = Write                 */
        char eof;                /* !=0 if end of file                  */
        };


 /** BF_Open : Opens the Bfile f
  **           Input  : the name of the Bfile f, the open mode (R or W)
  **           Output : a pointer to the Bfile if success, NULL
  **                       otherwise.
  **/
BFILE *BF_Open ( const char *Name, const char Mode);


 /** BF_Write : writes the nbits bits in the Bfile f
  **            Input  : the nbits bits, the Bfile f
  **            Output : Error if an error occured.
  **/
int BF_Write ( ulong bits, int nbits, BFILE *f);


 /** BF_Read : reads the nbits bits in the Bfile f
  **           Input  : the number nbits of bits to read, the Bfile f
  **           Output : the bits read.
  **/
ulong BF_Read ( int nbits, BFILE *f);



 /** BF_Close : Opens the bits Bfile f
  **            Input  : a pointer to the Bfile to close
  **            Output : Error if an error occured.
  **/
int BF_Close ( BFILE *f);




/***********************************
 **** Main functions definition ****
 ***********************************/

ulong Proba[256];   /* Array of probabilities of the 256 different bytes */
ulong In_Size;      /* The Input File Size    */
ulong Out_Size;     /* The Output File Size   */
byte  DICO_BL;      /* Dictionnary length in bits */

void Compress (BFILE *Input, BFILE *Output);

void Decompress (BFILE *Input, BFILE *Output);

void report(int Cmd, char *Input_Name);


# endif